"""
Decision Curve Analysis in Python

Modules exported by this package:

- `dca`
- `plot_graphs`
- `prevalence`
- `risks`

"""

__version__ = "1.0.6.4"


from dcurves import load_test_data
from dcurves.dca import dca
from dcurves.plot_graphs import plot_graphs
import os

data = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')



"""
This module houses plotting functions used in the user-facing plot_graphs() 
function to plot net-benefit scores and net interventions avoided.
"""
from typing import Optional, Iterable
import random
import matplotlib.pyplot as plt
import pandas as pd

def _get_colors(num_colors=None):
    """
    Generate a random tuple of colors of length num_colors

    Parameters
    ----------
    num_colors : int
        Number of colors to be outputted in tuple form

    Returns
    -------
    tuple
    """
    return [f"#{format(random.randint(0, 0xFFFFFF), '06x')}" for _ in range(num_colors)]


def _plot_net_benefit(
    plot_df: pd.DataFrame,
    y_limits: Iterable = (-0.05, 1),
    color_names: Iterable = None,
) -> None:
    """
    Plot net benefit values against threshold values.

    Parameters
    ----------
    plot_df : pd.DataFrame
        Data containing threshold values, model columns of net benefit scores
        to be plotted
    y_limits : list[float]
        2 floats, lower and upper bounds for y-axis
    color_names : list[str]
        Colors to render each model (if n models supplied, then need n+2 colors,
        since 'all' and 'none' models will be included by default

    Returns
    -------
    None
    """

    modelnames = plot_df["model"].value_counts().index
    for modelname, color_name in zip(modelnames, color_names):
        single_model_df = plot_df[plot_df["model"] == modelname]
        plt.plot(
            single_model_df["threshold"],
            single_model_df["net_benefit"],
            color=color_name,
        )
        plt.ylim(y_limits)
        plt.legend(modelnames)
        plt.grid(color='black', which="both", axis="both", linewidth="0.3")
        plt.xlabel("Threshold Values")
        plt.ylabel("Net Benefit")
    plt.show()


def _plot_net_intervention_avoided(
    plot_df: pd.DataFrame,
    y_limits: Iterable = (-0.05, 1),
    color_names: Iterable = None,
) -> None:
    """
    Plot net interventions avoided values against threshold values.

    Parameters
    ----------
    plot_df : pd.DataFrame
        Data containing threshold values, model columns of net intervention
        scores to be plotted
    y_limits : list[float]
        2 floats, lower and upper bounds for y-axis
    color_names
        Colors to render each model (if n models supplied, then need n+2 colors,
        since 'all' and 'none' models will be included by default

    Returns
    -------
    None
    """

    modelnames = plot_df["model"].value_counts().index
    for modelname, color_name in zip(modelnames, color_names):
        single_model_df = plot_df[plot_df["model"] == modelname]
        plt.plot(
            single_model_df["threshold"],
            single_model_df["net_intervention_avoided"],
            color=color_name,
        )

        plt.ylim(y_limits)
        plt.legend(modelnames)
        plt.grid(color='black', which="both", axis="both", linewidth="0.3")
        plt.xlabel("Threshold Values")
        plt.ylabel("Net Reduction of Interventions")
    plt.show()


def plot_graphs(
    plot_df: pd.DataFrame,
    graph_type: str = "net_benefit",
    y_limits: Iterable = (-0.05, 1),
    color_names: Optional[Iterable] = None,
) -> None:
    """
    Plot either net benefit or interventions avoided per threshold.

    Parameters
    ----------
    plot_df : pd.DataFrame
        Data containing threshold values, model columns of net benefit/intervention
        scores to be plotted
    graph_type : str
        Type of plot (either 'net_benefit' or 'net_intervention_avoided')
    y_limits : Iterable[Lower Bound, Upper Bound]
        2 floats, lower and upper bounds for y-axis
    color_names : Iterable[str]
        Colors to render each model (if n models supplied, then need n+2 colors,
        since 'all' and 'none' models will be included by default

    Returns
    -------
    None
    """

    if graph_type not in ["net_benefit", "net_intervention_avoided"]:
        raise ValueError(
            "graph_type must be one of 2 strings: net_benefit,"
            " net_intervention_avoided"
        )

    modelnames = plot_df["model"].value_counts().index
    if color_names is None:
        color_names = _get_colors(num_colors=len(modelnames))
    elif color_names is not None:
        if len(color_names) < len(modelnames):
            raise ValueError(
                "More predictors than color_names, please enter more color names"
                " in color_names list and try again"
            )

    if graph_type == "net_benefit":
        _plot_net_benefit(plot_df=plot_df, y_limits=y_limits, color_names=color_names)
    elif graph_type == "net_intervention_avoided":
        _plot_net_intervention_avoided(
            plot_df=plot_df, y_limits=y_limits, color_names=color_names
        )




"""
This module is commented out, and will be populated in the future to add checks
and display helpful error messages.
"""
# from .dca import *
#
#
# def _validate_time(time: float or int):
#     # make sure time input is a float or int
#
#     if time is None:
#         pass
#     elif isinstance(time, float) or isinstance(time, int):
#         pass
#     else:
#         raise TypeError("time must be of type: float or int")
#
#
# def _validate_time_to_outcome_col(time_to_outcome_col: str):
#     # make sure time_to_outcome_col input is a string
#
#     if time_to_outcome_col is None:
#         pass
#     elif isinstance(time_to_outcome_col, str):
#         pass
#     else:
#         raise TypeError("time_to_outcome_col must be of type: str")
#
#
# def _surv_convert_to_risk_input_checks(
#         time: float,
#         time_to_outcome_col: str):
#     # make sure _convert_to_risk inputs are kosher
#
#     _validate_time(time=time)
#     _validate_time_to_outcome_col(time_to_outcome_col=time_to_outcome_col)
#
#
# def _validate_thresholds(thresholds: list):
#     # make sure thresholds input is a list of floats
#
#     if not isinstance(thresholds, list):
#         raise TypeError('probabilities must be of type list')
#     else:
#         for i in thresholds:
#             if not isinstance(i, float):
#                 raise TypeError("time_to_outcome_col must be of type: float")
#         if not (len(thresholds) > 0):
#             raise ValueError("Thresholds must contain at least 1 value")
#
#
# def _binary_calculate_test_consequences_input_checks(
#         thresholds: list):
#     # make sure _calculate_test_consequences inputs are kosher
#
#     _validate_thresholds(thresholds=thresholds)
#
#
# def _surv_calculate_test_consequences_input_checks(
#         thresholds: list,
#         time: float,
#         time_to_outcome_col: str):
#     # make sure _calculate_test_consequences inputs are kosher
#
#     _validate_thresholds(thresholds=thresholds)
#     _validate_time(time=time)
#     _validate_time_to_outcome_col(time_to_outcome_col=time_to_outcome_col)
#
# def _validate_str_list(input_list: list):
#     # Make sure input is a list, and all list elements are strings
#
#     if input_list is None:
#         pass
#     elif not isinstance(input_list, list):
#         raise TypeError('Predictors must be of type list')
#     else:
#         for i in input_list:
#             if not isinstance(i, str):
#                 raise TypeError('All elements in predictors list must be of type str')
#
#
# def _validate_thresh_list(input_list: list):
#     # make sure input_list input is a list, and all list elements are int/float
#
#     if not isinstance(input_list, list):
#         raise TypeError('input must be of type list')
#     else:
#         for i in input_list:
#             if not (isinstance(i, float) or isinstance(i, int)):
#                 raise TypeError("input list contents must be of type: float/int")
#
#     if not (len(input_list) == 3):
#         raise ValueError("input list must contain 3 values")
#
# def _validate_limits_list(input_list: list):
#     # make sure input_list input is a list, and all list elements are int/float
#
#     if not isinstance(input_list, list):
#         raise TypeError('input must be of type list')
#     else:
#         for i in input_list:
#             if not (isinstance(i, float) or isinstance(i, int)):
#                 raise TypeError("input list contents must be of type: float/int")
#
#     if not (len(input_list) == 2):
#         raise ValueError("input list must contain 2 values")
#
#
# def _validate_harm(harm: dict):
#     # make sure harm input is a dict
#
#     if harm is None:
#         pass
#     elif not isinstance(harm, dict):
#         raise TypeError('harm input not of type dict')
#
#
# def _validate_probabilities(probabilities: list, predictors: list):
#     # make sure probabilities input is a list, the length of
#     # probabilities/predictors is the same, and all elements
#     # in probabilities is of type bool
#
#     _validate_str_list(input_list=predictors)
#
#     if probabilities is None:
#         pass
#     elif not isinstance(probabilities, list):
#         raise TypeError('probabilities must be of type list')
#     else:
#         if not (len(probabilities) == len(predictors)):
#             raise ValueError('probabilities input and predictors input must be of same length')
#         for i in probabilities:
#             if not isinstance(i, bool):
#                 raise TypeError('All elements in probabilities must be of type bool')
#
#
# def _binary_dca_input_checks(
#         predictors: list,
#         thresh_vals: list,
#         harm: dict,
#         probabilities: list,  # list of TRUE/FALSE values indicating which predictors
#         ):
#     # make sure dca inputs are kosher
#
#     _validate_str_list(input_list=predictors)
#     _validate_thresh_list(input_list=thresh_vals)
#     _validate_harm(harm=harm)
#     _validate_probabilities(probabilities=probabilities, predictors=predictors)
#
# def _surv_dca_input_checks(
#         predictors: list,
#         thresh_vals: list,
#         harm: dict,
#         probabilities: list,  # list of TRUE/FALSE values indicating which predictors
#         time: float,
#         prevalence: float,
#         time_to_outcome_col: str):
#     # make sure dca inputs are kosher
#
#     _validate_str_list(input_list=predictors)
#     _validate_thresh_list(input_list=thresh_vals)
#     _validate_harm(harm=harm)
#     _validate_probabilities(probabilities=probabilities, predictors=predictors)
#     _validate_time(time=time)
#     _validate_time_to_outcome_col(time_to_outcome_col=time_to_outcome_col)
#
#
# def _validate_graph_type(graph_type: str):
#
#     if not isinstance(graph_type, str):
#         raise TypeError('input must be of type str')
#
#     if graph_type not in ['net_benefit', 'net_intervention_avoided']:
#         raise ValueError('graph_type must be either: net_benefit or net_intervention_avoided')
#
#
# def _plot_graphs_input_checks(
#         y_limits: list,
#         color_names: list,
#         graph_type: str):
#     # make sure output_df input is pandas DataFrame
#
#     _validate_limits_list(input_list=y_limits)
#     _validate_str_list(input_list=color_names)
#     _validate_graph_type(graph_type=graph_type)
#
# def _plot_net_intervention_input_checks(
#         after_dca_df: pd.DataFrame,
#         graph_type: str):
#
#     if graph_type == 'net_intervention_avoided':
#         if 'net_intervention_avoided' not in after_dca_df.columns:
#             raise ValueError("""net_intervention_avoided field not found in inputted dataframe
#             columns - must run net_intervention_avoided() function to calculate net
#             interventions avoided, and reinput returned dataframe into plotting function""")



# """
# This module houses user-facing simulation data-retrieval functions for end-users.
# These data were also used for testing.
# """
# from importlib.resources import files
# import pandas as pd
# from dcurves import data

# def load_binary_df():
#     """
#     Load Simulation Data For Binary Endpoints DCA
#     :return pd.DataFrame that contains binary data
#     """
#     file_path = files(data).joinpath("df_binary.csv")
#     return pd.read_csv(file_path)

# def load_survival_df():
#     """
#     Load Simulation Data For Survival Endpoints DCA
#     :return pd.DataFrame that contains survival data
#     """
#     file_path = files(data).joinpath("df_surv.csv")
#     return pd.read_csv(file_path)


# def load_case_control_df():
#     """
#     Load Simulation Data For Binary Endpoints DCA With User-Specified Outcome
#     Prevalence (Case-Control)
#     :return pd.DataFrame that contains binary data
#     """
#     file_path = files(data).joinpath("df_case_control.csv")
#     return pd.read_csv(file_path)

"""
This module houses user-facing simulation data-retrieval functions for end-users.
These data were also used for testing.
"""
import os
import pandas as pd

def _load_data_from_local(filename):
    """
    Load data from the local 'data' directory.
    :param filename: Name of the file to be loaded.
    :return: pd.DataFrame containing the data from the file.
    """
    # Assuming data directory is in the same level as this file
    data_dir = os.path.join(os.path.dirname(__file__), 'data')
    file_path = os.path.join(data_dir, filename)
    return pd.read_csv(file_path)

def load_binary_df():
    """
    Load Simulation Data For Binary Endpoints DCA
    :return pd.DataFrame that contains binary data
    """
    return _load_data_from_local('df_binary.csv')

def load_survival_df():
    """
    Load Simulation Data For Survival Endpoints DCA
    :return pd.DataFrame that contains survival data
    """
    return _load_data_from_local('df_surv.csv')

def load_case_control_df():
    """
    Load Simulation Data For Binary Endpoints DCA With User-Specified Outcome
    Prevalence (Case-Control)
    :return pd.DataFrame that contains binary data
    """
    return _load_data_from_local('df_case_control.csv')



"""
This module houses the functions used to calculate risk scores for
binary and survival endpoints. It is kept in a separate module
from dca.py to disperse the code and make clear its individual
dependencies.
"""
import sys
from typing import Optional, Union
import pandas as pd
import statsmodels.api as sm
import lifelines


def _calc_binary_risks(data: pd.DataFrame, outcome: str, model: str) -> list:
    """
    Calculate Risks For a Model Column for binary DCA.

    Parameters
    ----------
    data : pd.DataFrame
        Initial raw data ideally containing risk scores (scores ranging from 0 to 1),
        or else predictor/model values, and outcome of interest
    outcome : str
        Column name of outcome of interest in risks_df
    model
        Model column name in risks_df

    Returns
    -------
    list[float]
        list of predicted risk scores for a model column
    """
    predicted_vals = (
        sm.formula.glm(outcome + "~" + model, family=sm.families.Binomial(), data=data)
        .fit()
        .predict()
    )
    return list(predicted_vals)


def _calc_surv_risks(
    data: pd.DataFrame,
    outcome: str,
    model: str,
    time: Union[int, float],
    time_to_outcome_col: str,
) -> list:
    """
    Calculate Risks For a Model Column for survival DCA.

    Parameters
    ----------
    data : pd.DataFrame
        Initial raw data ideally containing risk scores (scores ranging from 0 to 1),
        or else predictor/model values, and outcome of interest
    outcome : str
        Column name of outcome of interest in risks_df
    model : str
        Model column name in risks_df
    time : int or float
        Time of interest in years, used in Survival DCA
    time_to_outcome_col : str
        Column name containing time to outcome values, used in Survival DCA

    Returns
    -------
    list
        list of coxph-predicted risk scores for a model column

    """
    cph_df = data[[time_to_outcome_col, outcome, model]].copy()
    cph = lifelines.CoxPHFitter()
    cph.fit(cph_df, time_to_outcome_col, outcome)
    cph_df = cph_df.assign(**{time_to_outcome_col: time}).copy()
    predicted_vals = cph.predict_survival_function(cph_df, times=time).values[0]
    return list(predicted_vals)


def _create_risks_df(
    data: pd.DataFrame,
    outcome: str,
    models_to_prob: Optional[list] = None,
    time: Optional[Union[float, int]] = None,
    time_to_outcome_col: Optional[str] = None,
) -> pd.DataFrame:
    """
    Parameters
    ----------
    data : pd.DataFrame
        Initial raw data ideally containing risk scores (scores ranging from 0 to 1),
        or else predictor/model values, and outcome of interest
    outcome : str
        Column name of outcome of interest in risks_df
    models_to_prob : list[str]
        Columns that need to be converted to risk scores from 0 to 1
    time : int or float
        Time of interest in years, used in Survival DCA
    time_to_outcome_col : str
        Column name in data containing time to outcome values, used in Survival DCA

    Returns
    -------
    pd.DataFrame
        Data with non-risk score model columns converted to risk scores between 0 and 1
    """
    if models_to_prob is None:
        pass
    elif time_to_outcome_col is None:
        for model in models_to_prob:
            binary_risks = _calc_binary_risks(data=data, outcome=outcome, model=model)

            data = data.assign(**{model: binary_risks}).copy()

    elif time_to_outcome_col is not None:
        for model in models_to_prob:
            surv_risks = _calc_surv_risks(
                data=data,
                outcome=outcome,
                model=model,
                time=time,
                time_to_outcome_col=time_to_outcome_col,
            )
            data = data.assign(**{model: surv_risks}).copy()

    data = data.assign(**{"all": 1}).copy()
    data = data.assign(**{"none": 0}).copy()

    return data


def _rectify_model_risk_boundaries(
    risks_df: pd.DataFrame, modelnames: list
) -> pd.DataFrame:
    """
    Parameters
    ----------
    risks_df : pd.DataFrame
        Data containing (converted if necessary) risk scores (scores ranging from 0 to 1)
        and outcome of interest
    modelnames : list[str]
        Column names from risks_df that contain model risk scores

    Returns
    -------
    pd.DataFrame
        Data with model column risk scores 0 and 1 changed to 0 - e and 1 + e, respectively,
        to evaluate as less/greater than 0 and 1 thresholds for correct tp/fp evaluations
    """

    machine_epsilon = sys.float_info.epsilon

    modelnames = modelnames + ["all", "none"]
    for modelname in modelnames:
        risks_df[modelname].replace(
            to_replace=0, value=0 - machine_epsilon, inplace=True
        )
        risks_df[modelname].replace(
            to_replace=1, value=1 + machine_epsilon, inplace=True
        )

    return risks_df



"""
This module houses the functions used to calculate prevalence. It is
kept in a separate module from dca.py to disperse the code and make
clear its individual dependencies.
"""

from typing import Optional, Union
import pandas as pd
import lifelines


def _calc_prevalence(
    risks_df: pd.DataFrame,
    outcome: str,
    prevalence: Optional[Union[int, float]] = None,
    time: Optional[Union[int, float]] = None,
    time_to_outcome_col: Optional[str] = None,
) -> float:
    """
    Calculate prevalence value when not supplied for binary and survival DCA cases,
    and set prevalence value for binary case when supplied (case control). Case
    control prevalence is not value for survival cases.

    Parameters
    ----------
    risks_df : pd.DataFrame
        Data containing (converted if necessary) risk scores (scores ranging from
        0 to 1) and outcome of interest
    outcome : str
        Column name of outcome of interest in risks_df
    prevalence : int or float
        Value that indicates the prevalence among the population, only to be
        specified in case-control situations
    time : int or float
        Time of interest in years, used in Survival DCA
    time_to_outcome_col : str
        Column name of time of interest in risks_df

    Returns
    -------
    float
        Either calculated prevalence or inputted prevalence depending on
        whether input prevalence was supplied
    """

    if time_to_outcome_col is None:
        if prevalence is None:
            prevalence = sum(risks_df[outcome]) / len(risks_df[outcome])
    else:
        if prevalence is not None:
            raise ValueError("In survival outcomes, prevalence should not be supplied")

        kmf = lifelines.KaplanMeierFitter()
        kmf.fit(risks_df[time_to_outcome_col], risks_df[outcome] * 1)
        prevalence = 1 - kmf.survival_function_at_times(time).iloc[0]
    return float(prevalence)
